# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Common utility functions used throughout the ``hips`` package.

Note that those are mostly for internal notes, if you use them from
external scripts or packages at all, you should not expect stability
from our side and it might be better if you copy over what you need.
"""
