"""
This package contains utilities and extensions for the Astropy sphinx
documentation.  In particular, the `astropy.sphinx.conf` should be imported by
the sphinx ``conf.py`` file for affiliated packages that wish to make use of
the Astropy documentation format. Note that some sphinx extensions which are
bundled as-is (numpydoc and sphinx-automodapi) are included in
astropy_helpers.extern rather than astropy_helpers.sphinx.ext.
"""
